/* ==========================================================================*/
/* - - - - - - - - - - - - -   G p s . c   - - - - - - - - - - - - - - - - - */
/* ==========================================================================*/
/*    Implementation File for the gps module                                 */
/*	Created   2011-04-23   M. Wheatley                                       */
/*	Modified  2011-04-23   MW                                                */
/* WSPRbeacon   AT91SAM7256 gps uart handler                                */
///////////////////////////////////////////////////////////////////////////////
//            2011 Copyright (C) Moe Wheatley, AE4JY
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//////////////////////////////////////////////////////////////////////////////

/*---------------------------------------------------------------------------*/
/*------------------------> I N C L U D E S <--------------------------------*/
/*---------------------------------------------------------------------------*/
#include "board.h"
#include "gps.h"
#include "irq.h"
#include "term.h"
#include "wspr.h"
#include "iicctrl.h"
#include "timer.h"
#include "protocoldefs.h"
#include <stdlib.h>
#include "string.h"
#include "wsprencoder.h"
 
/*---------------------------------------------------------------------------*/
/*--------------------> L O C A L   D E F I N E S <--------------------------*/
/*---------------------------------------------------------------------------*/
//* Standard Asynchronous Mode : 8 bits , 1 stop , no parity
#define GPSASYNC_MODE ( AT91C_US_USMODE_NORMAL + \
                        AT91C_US_NBSTOP_1_BIT + \
                        AT91C_US_PAR_NONE + \
                        AT91C_US_CHRL_8_BITS + \
                        AT91C_US_CLKS_CLOCK )

#define GPS_BAUD 4800

#define STATE_TOP 0
#define STATE_SENTENCE 1
#define STATE_TIME 2
#define STATE_LAT 3
#define STATE_LATH 4
#define STATE_LON 5
#define STATE_LONH 6
#define STATE_STATUS 7
#define STATE_SATS 8
#define STATE_HDIL 9
#define STATE_ALT 10



#define KEEPALIVE 50


#define FREERUN_TIMELIMIT (60*60*48)	//time allowed in seconds to run off internal clock



/*---------------------------------------------------------------------------*/
/*-------------------> M E M O R Y  A L L O C A T I O N <--------------------*/
/*---------------------------------------------------------------------------*/
uint8_t GpsRxQueue[GPSQSIZE];
volatile uint32_t GpsRxHead;
uint8_t GpsStatus;
static uint32_t GpsRxTail;
xSemaphoreHandle xGpsRxSemaphore = NULL;
BOOL TimeIsAccurate;
BOOL GpsTimeIsValid;
uint8_t GPSHour;
uint8_t GPSMinute;
uint8_t GPSSecond;
uint8_t LatStr[MAX_PARAM];
uint8_t LonStr[MAX_PARAM];
uint32_t Latitude;	//0 to 180 deg  times 100000
uint32_t Longitude;	//0 to 360 deg  times 100000
uint8_t NumSats;
uint8_t MHStr[7];

static uint32_t FreeRunTimer;
static uint32_t MsgState;
static uint32_t BCount;
static uint8_t Sentence[MAX_PARAM];
static uint8_t SatStatus;
static uint8_t TimeStr[MAX_PARAM];

static char NumSatsStr[MAX_PARAM];
static char AltitudeStr[MAX_PARAM];


/*---------------------------------------------------------------------------*/
/*--------------------------> P R O T O T Y P E S <--------------------------*/
/*---------------------------------------------------------------------------*/
static void GpsTask( void *pvParmeters );
static void ManageStatus( void );
static BOOL ParseNMEAByte(uint8_t ch);
static void CalcGridSquare(uint8_t* pStr);	//fills in Grid square string


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                        ----------------                                   */
/*                       |  G p s I n i t |                                  */
/*                        ----------------                                   */
/*    Called to initialize the GPS USART 0 port.                             */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void GpsInit( )
{
	AT91F_PMC_EnablePeriphClock ( AT91C_BASE_PMC, 1 << AT91C_ID_US1 ) ;
 	//* Configure uart PIO controllers to periph mode
	AT91F_PIO_CfgPeriph( AT91C_BASE_PIOA,
			((unsigned int) GPSTX   ) |
			((unsigned int) GPSRX   ),	// Peripheral A
			((unsigned int)	PPS)	); 	// Peripheral B
	// Usart Configure
	AT91C_BASE_US1->US_CR = AT91C_US_RSTRX | AT91C_US_RSTTX;
	AT91C_BASE_US1->US_CR = AT91C_US_RSTSTA;
	AT91C_BASE_US1->US_BRGR = ((MCK*10)/(16*GPS_BAUD) + 5) /10;	//use nearest integer;
	AT91C_BASE_US1->US_MR = GPSASYNC_MODE;
	AT91C_BASE_US1->US_TTGR = 0;	//no inter tx spacing
	AT91C_BASE_US1->US_RTOR = 0xFFFF;
	// Enable usart
	AT91C_BASE_US1->US_CR = AT91C_US_RXEN|AT91C_US_TXEN;
 	//* Enable USART irqs
	AT91F_US_EnableIt(AT91C_BASE_US1, AT91C_US_FRAME | AT91C_US_OVRE |AT91C_US_RXRDY);
	//* Config Usart interrupt
	AT91F_AIC_ConfigureIt ( AT91C_BASE_AIC, AT91C_ID_US1, USART1_INTERRUPT_LEVEL,
			AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL, GpsIrqService);
	AT91F_AIC_EnableIt (AT91C_BASE_AIC, AT91C_ID_US1);

	xTaskCreate( GpsTask, ( const signed portCHAR * const ) "GPS",
				configMINIMAL_STACK_SIZE, NULL, GPS_TASK_PRIORITY, ( xTaskHandle * ) NULL );
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      ---------------                                      */
/*                     | G p s T a s k |                                     */
/*                      ---------------                                      */
/*    Task that process GPS usart rx bytes                                   */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
static void GpsTask( void *pvParmeters )
{
uint8_t ch;
	vTaskDelay(3000);
	MsgState = STATE_TOP;
	BCount = 0;
	GpsRxHead = 0;
	GpsRxTail = 0;
	GPSHour = 0;
	GPSMinute = 0;
	GPSSecond = 0;
	GpsStatus = GPS_STATUS_NO_SAT;
	NumSatsStr[0] = 0;
	AltitudeStr[0] = 0;
	LatStr[0] = 0;
	LonStr[0] = 0;
	NumSats = 0;
	FreeRunTimer = 0;
	TimeIsAccurate = FALSE;
	GpsTimeIsValid = FALSE;
	vSemaphoreCreateBinary( xGpsRxSemaphore );
	xSemaphoreTake( xGpsRxSemaphore, ( portTickType )10 );
	if( 0 == (AT91F_PIO_GetInput(AT91C_BASE_PIOA) & GPSON) )
	{
		AT91F_PIO_SetOutput( AT91C_BASE_PIOA, GPSTOG );
		vTaskDelay(100);
		AT91F_PIO_ClearOutput( AT91C_BASE_PIOA, GPSTOG );
	}
	ManageStatus();
	for( ;; )	//loop forever in this task
    {
		if( xSemaphoreTake( xGpsRxSemaphore, ( portTickType )100 ) == pdTRUE )
		{
			while( GpsRxHead != GpsRxTail)
			{
				ch = GpsRxQueue[GpsRxTail++];
				if(CI_OPMODE_GPSRAW == Settings.OpMode)
					SendTermUart(ch);
				if(GpsRxTail>=GPSQSIZE)
					GpsRxTail = 0;
				if( ParseNMEAByte(ch) )
				{	//got complete GPS NMEA msg
					NumSats = atoi(NumSatsStr);
					ManageStatus();
				}
			}
		}
		else	//nothing coming in from GPS serial port(probably not awake)
		{
			if( 0 == (AT91F_PIO_GetInput(AT91C_BASE_PIOA) & GPSON) )
			{
				GpsStatus = GPS_STATUS_NO_SAT;
				AT91F_PIO_SetOutput( AT91C_BASE_PIOA, GPSTOG );
				vTaskDelay(100);
				AT91F_PIO_ClearOutput( AT91C_BASE_PIOA, GPSTOG );
			}
		}
	}
}



/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      -------------------------------                      */
/*                     | M a n a g e G p s S t a t u s |                     */
/*                      -------------------------------                      */
/* Called to manage the GPS status states                                    */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
static void ManageStatus( void )
{
//commented out for passive antenna
//	if( AT91F_PIO_GetInput(AT91C_BASE_PIOA) & ANTOC )
//		GpsStatus = GPS_STATUS_ERROR_ANT;
	if(FreeRunTimer)
	{
		FreeRunTimer--;
//		printf("FRT %d\r\n",FreeRunTimer);
		if(!FreeRunTimer)
		{
			TimeIsAccurate = FALSE;
		}
	}
	switch(GpsStatus)
	{
		case GPS_STATUS_ERROR_ANT:
//			printf("Antenna Disconnected\r\n");
//			if( !(AT91F_PIO_GetInput(AT91C_BASE_PIOA) & ANTOC) )
				GpsStatus = GPS_STATUS_NO_SAT;
			GpsTimeIsValid = FALSE;
			break;
		case GPS_STATUS_NO_SAT:
			if(CI_OPMODE_GPS == Settings.OpMode)
				printf("\r\nNo Satellites\r\n");
			if( (NumSats>0) && (SatStatus!=0) )
				GpsStatus = GPS_STATUS_TIMESYNC;
			GpsTimeIsValid = FALSE;
			break;
		case GPS_STATUS_TIMESYNC:
			if(CI_OPMODE_GPS == Settings.OpMode)
				printf("\r\nTime Sync  NumSats = %d\r\n",NumSats);
			if( (NumSats==0) || (SatStatus==0))
				GpsStatus = GPS_STATUS_NO_SAT;
			else if( (PpsPeriod>23999000) && (PpsPeriod<24000100)  )
				GpsStatus = GPS_STATUS_FULLSYNC;
			GpsTimeIsValid = FALSE;
			break;
		case GPS_STATUS_FULLSYNC:
			if( (NumSats<3) || (SatStatus==0))
			{
				GpsStatus = GPS_STATUS_NO_SAT;
			}
			else if( (PpsPeriod<23999000) || (PpsPeriod>24000100)  )
			{
				GpsStatus = GPS_STATUS_TIMESYNC;
			}
			else
			{	//extract UTC time
				GPSHour = (TimeStr[0]-'0')*10 + (TimeStr[1]-'0');
				GPSMinute = (TimeStr[2]-'0')*10 + (TimeStr[3]-'0');
				GPSSecond = (TimeStr[4]-'0')*10 + (TimeStr[5]-'0');
				NumSats = atoi(NumSatsStr);	//get number of Sats in view
				CalcGridSquare(MHStr);	//create maidenhead gridsquare
				FreeRunTimer = FREERUN_TIMELIMIT;	//reset free running timer
				TimeIsAccurate = TRUE;
				GpsTimeIsValid = TRUE;
				if(CI_OPMODE_GPS == Settings.OpMode)
				{
					printf("\r\nNumSats=%d   Grid Square %s\r\n",NumSats, MHStr);
					printf("UTC %d Hours : %d Minutes : %d Seconds\r\n",GPSHour, GPSMinute, GPSSecond);
					printf("Latitude %s   Longitude %s\r\n",LatStr, LonStr);
					printf("Altitude %s meters\r\n",AltitudeStr);
				}
			}
			break;
		default:
			GpsStatus = GPS_STATUS_NO_SAT;
			break;
	}
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                    -----------------------------                          */
/*                   | C a l c G r i d S q u a r e |                         */
/*                    -----------------------------                          */
/* Calculates 6 character Maidenhead grid square from current Lat/Lon        */
/*  Lat and Lon are in degrees * 100000  to keep fixed point math.           */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void CalcGridSquare(uint8_t* pStr)
{
uint32_t tmp;
	//preload MH string with base characters
	pStr[0] = 'A';pStr[1] = 'A'; pStr[2] = '0'; pStr[3] = '0';
	pStr[4] = 'a';   pStr[5] = 'a'; pStr[6] = 0;

	//convert GPS ASCII lat lon into integer values multiplied by 100000
	Longitude = (LonStr[1]-'0')*10000000 + (LonStr[2]-'0')*1000000 + (LonStr[3]-'0')*100000 ;
	tmp = (LonStr[4]-'0')*1000000 + (LonStr[5]-'0')*100000 +
			(LonStr[7]-'0')*10000 + (LonStr[8]-'0')*1000 + (LonStr[9]-'0')*100 + (LonStr[10]-'0')*10;
	Longitude += (tmp/60);
	Latitude = (LatStr[1]-'0')*1000000 + (LatStr[2]-'0')*100000 ;
	tmp = (LatStr[3]-'0')*1000000 + (LatStr[4]-'0')*100000 +
			(LatStr[6]-'0')*10000 + (LatStr[7]-'0')*1000 + (LatStr[8]-'0')*100 + (LatStr[9]-'0')*10;
	Latitude += (tmp/60);

	//convert to all positive MH degrees lat 0 to 90 and lon 0 to 180
	if('W' == LonStr[0])
		Longitude = 18000000 - Longitude;
	else
		Longitude = Longitude + 18000000;
	if('N' == LatStr[0])
		Latitude = Latitude + 9000000;
	else
		Latitude = 900000 - Latitude;

	//now create MH grid square string
	tmp = Longitude/20;
	pStr[0] += (tmp/100000);
	tmp = (tmp - (tmp/100000)*100000) * 10;
	pStr[2] += (tmp/100000);
	tmp = (tmp - (tmp/100000)*100000) * 24;
	pStr[4] += (tmp/100000);

	tmp = Latitude/10;
	pStr[1] += (tmp/100000);
	tmp = (tmp - (tmp/100000)*100000) * 10;
	pStr[3] += (tmp/100000);
	tmp = (tmp - (tmp/100000)*100000) * 24;
	pStr[5] += (tmp/100000);
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                    ---------------------------                            */
/*                   | P a r s e N M E A B y t e |                           */
/*                    ---------------------------                            */
/*    This routine is called to parse a new message from the Gps.            */
/* Returns TRUE if '$GGA' message is decoded.                                */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
static BOOL ParseNMEAByte(uint8_t ch)
{
BOOL ret = FALSE;
	if( ('$' == ch) || (BCount>=MAX_PARAM) )
		MsgState = STATE_TOP;
	switch(MsgState)
	{
		case STATE_TOP:
			BCount = 0;
			if( '$' == ch)
				MsgState = STATE_SENTENCE;
			break;
		case STATE_SENTENCE:
			if( ',' == ch )
			{
				BCount = 0;
				if( ('G'==Sentence[2]) && ('G'==Sentence[3]) && ('A'==Sentence[4]) )
					MsgState = STATE_TIME;
				else
					MsgState = STATE_TOP;
			}
			else
				Sentence[BCount++] = ch;
		break;
		case STATE_TIME:
			if( ',' == ch )
			{
				TimeStr[BCount] = 0;
				BCount = 1;
				MsgState = STATE_LAT;
			}
			else
				TimeStr[BCount++] = ch;
			break;
		case STATE_LAT:
			if( ',' == ch )
				MsgState = STATE_LATH;
			else
				LatStr[BCount++] = ch;
			break;
		case STATE_LATH:
			if( ',' == ch )
			{
				LatStr[BCount] = 0;
				BCount = 1;
				MsgState = STATE_LON;
			}
			else
				LatStr[0] = ch;
			break;
		case STATE_LON:
			if( ',' == ch )
				MsgState = STATE_LONH;
			else
				LonStr[BCount++] = ch;
			break;
		case STATE_LONH:
			if( ',' == ch )
			{
				LonStr[BCount] = 0;
				BCount = 0;
				MsgState = STATE_STATUS;
			}
			else
				LonStr[0] = ch;
			break;
		case STATE_STATUS:
			if( ',' == ch )
			{
				BCount = 0;
				MsgState = STATE_SATS;
			}
			else
				SatStatus = ch-'0';
			break;
		case STATE_SATS:
			if( ',' == ch )
			{
				NumSatsStr[BCount++] = 0;
				BCount = 0;
				MsgState = STATE_HDIL;
			}
			else
				NumSatsStr[BCount++] = ch;
			break;
		case STATE_HDIL:
			if( ',' == ch )
			{
				BCount = 0;
				MsgState = STATE_ALT;
			}
			break;
		case STATE_ALT:
			if( ',' == ch )
			{
				AltitudeStr[BCount++] = 0;
				BCount = 0;
				MsgState = STATE_TOP;
				ret = TRUE;
			}
			else
				AltitudeStr[BCount++] = ch;
			break;
		default:
			BCount = 0;
			MsgState = STATE_TOP;
			break;
	}
	return ret;
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      ---------------------                                */
/*                     | S e n d G p s S t r |                               */
/*                      ---------------------                                */
/* Called to send string to the usart. (waits for previous byte to finish)   */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void SendGpsStr( char* pStr)
{
	while(0 != *pStr)
	{
		SendGpsUart(*pStr++);
	}
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                      -----------------------                              */
/*                     | S e n d G p s U a r t |                             */
/*                      -----------------------                              */
/* Called to send one byte out the usart. (waits for previous byte to finish)*/
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void SendGpsUart(unsigned char ch)
{
unsigned int status;
	do
	{
		status = *AT91C_US1_CSR;
	}while( 0 == (status & AT91C_US_TXRDY) );
	*AT91C_US1_THR = ch;
}
