/* ==========================================================================*/
/* - - - - - - - - - - - - - -  M a i n . C   - - - - - - - - - - - - - - - -*/
/* ==========================================================================*/
/*	Created   2011-09-21   M. Wheatley                                       */
/*	Modified  2011-09-21   MW                                                */
/*.........................................................................  */
/* WSPR Beacon  AT91SAM7256 Firmware                                         */
///////////////////////////////////////////////////////////////////////////////
//            2011 Copyright (C) Moe Wheatley, AE4JY
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//////////////////////////////////////////////////////////////////////////////

/*---------------------------------------------------------------------------*/
/*------------------------> I N C L U D E S <--------------------------------*/
/*---------------------------------------------------------------------------*/
#include "board.h"
#include "common.h"
#include "usart.h"
#include "timer.h"
#include "si5338.h"
#include "irq.h"
#include "gps.h"
#include "term.h"
#include "wspr.h"
#include "iicctrl.h"
#include "adctrl.h"
#include <stdlib.h>


/* Scheduler includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"


/*---------------------------------------------------------------------------*/
/*--------------------> L O C A L   D E F I N E S <--------------------------*/
/*---------------------------------------------------------------------------*/
#define QUEUE_SIZE 3

/*---------------------------------------------------------------------------*/
/*-------------------> M E M O R Y  A L L O C A T I O N <--------------------*/
/*---------------------------------------------------------------------------*/
 
/*---------------------------------------------------------------------------*/
/*------------------> L O C A L    P R O T O T Y P E S <---------------------*/
/*---------------------------------------------------------------------------*/
void LedTask( void *pvParmeters );

/*---------------------------------------------------------------------------*/
/*-------------------> M E M O R Y  A L L O C A T I O N <--------------------*/
/*---------------------------------------------------------------------------*/

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                             ---------                                     */
/*                            | M a i n |                                    */
/*                             ---------                                     */
/*    This is the program entry point.                                       */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
int main( void )
{
	AT91C_BASE_AIC->AIC_EOICR = 0;
    // enable the clock of the PIO
	AT91F_PMC_EnablePeriphClock ( AT91C_BASE_PMC, 1 << AT91C_ID_PIOA );
	AT91F_PIO_CfgPullup(AT91C_BASE_PIOA, ~(GLED|YLED|RLED|ANTOC|GPSON|PPS));	//no int pullups
	AT91F_PIO_CfgOpendrain(AT91C_BASE_PIOA,TXOFF);
	AT91F_PIO_SetOutput( AT91C_BASE_PIOA, TXOFF);
	AT91F_PIO_ClearOutput( AT91C_BASE_PIOA, GLED|RLED|YLED|GPSTOG|TP1|TP2);
	AT91F_PIO_CfgOutput( AT91C_BASE_PIOA, GLED|RLED|YLED|GPSTOG|TP1|TP2|TXOFF );
	AT91F_PIO_CfgInput( AT91C_BASE_PIOA, SW1|SW2|SW4|SW8|ANTOC|GPSON|PPS );

	ADCtrlInit();
	IICCtrlInit();
	TermInit();
	TimerInit();
	UsartInit(UART_BAUD);
	GpsInit();
	WsprInit();

	xTaskCreate( LedTask, ( const signed portCHAR * const ) "LED", 
					configMINIMAL_STACK_SIZE, NULL, LED_TASK_PRIORITY, ( xTaskHandle * ) NULL );
	vTaskStartScheduler();
	return 0;	//never gets here after RTOS started
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                             ---------------                               */
/*                            | L e d T a s k |                              */
/*                             ---------------                               */
/*    This tasks manages the status LED display.                             */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void LedTask( void *pvParmeters )
{
	SetYLedPWM( LED_ON );
	SetGLedPWM( LED_ON );
	SetRLedPWM( LED_ON );
	vTaskDelay(100);
	SetGLedPWM( LED_OFF );
	SetRLedPWM( LED_OFF );
	Si5338Init();
	for( ;; )
	{
		switch(	Settings.OpMode)
		{
			case CI_OPMODE_SETUP:
				SetYLedPWM( LED_OFF );
				SetRLedPWM( LED_OFF );
				SetGLedPWM( LED_ON );
				vTaskDelay(900);
				SetGLedPWM( LED_OFF );
				vTaskDelay(100);
				break;
			case CI_OPMODE_BOOT:
				SetYLedPWM( LED_ON );
				SetGLedPWM( LED_ON );
				SetRLedPWM( LED_ON );
				vTaskDelay(50);
				SetGLedPWM( LED_OFF );
				SetRLedPWM( LED_OFF );
				SetYLedPWM( LED_OFF );
				vTaskDelay(50);
				break;
			case CI_OPMODE_TUNE:
				//Red and Yellow controlled by VSWR pwm logic
				SetGLedPWM( LED_ON );
				vTaskDelay(10);
				SetGLedPWM( LED_OFF );
				vTaskDelay(990);
				break;
			case CI_OPMODE_GPS:
				SetGLedPWM( LED_ON );
				vTaskDelay(100);
				SetGLedPWM( LED_OFF );
				SetYLedPWM( LED_ON );
				vTaskDelay(100);
				SetYLedPWM( LED_OFF );
				SetRLedPWM( LED_ON );
				vTaskDelay(100);
				SetRLedPWM( LED_OFF );
				SetYLedPWM( LED_ON );
				vTaskDelay(100);
				SetYLedPWM( LED_OFF );
				break;
			case CI_OPMODE_GPSRAW:
				SetYLedPWM( LED_ON );
				SetGLedPWM( LED_ON );
				SetRLedPWM( LED_ON );
				vTaskDelay(900);
				SetGLedPWM( LED_OFF );
				SetRLedPWM( LED_OFF );
				SetYLedPWM( LED_OFF );
				vTaskDelay(100);
				break;
			case CI_OPMODE_WSPR:
				if(TimeIsAccurate)
					SetGLedPWM( LED_ON );
				else
					SetGLedPWM( LED_OFF );
				if(	GPS_STATUS_FULLSYNC == GpsStatus )
				{
					AT91F_PIO_ClearOutput( AT91C_BASE_PIOA, YLED );
				}
				else if(GPS_STATUS_TIMESYNC == GpsStatus)
				{
					SetYLedPWM( LED_ON );
					vTaskDelay(100);
					SetYLedPWM( LED_OFF );
				}
				else
				{
					AT91F_PIO_SetOutput( AT91C_BASE_PIOA, YLED );
				}
				if(!RfOnState)
				{
					if(TxError)
					{
						SetRLedPWM( LED_ON );
						vTaskDelay(50);
						SetRLedPWM( LED_OFF );
						vTaskDelay(50);
					}
					else
					{
						SetRLedPWM( LED_OFF );
						vTaskDelay(900);
					}
				}
				else
					vTaskDelay(900);
				break;
		}
//
		//manage any pending EEROM writes
		if(NeedToSaveSettings)
		{
			NeedToSaveSettings = FALSE;
			SaveAll();
		}
		rand();		//keep randomizer running
    }
}

